from subprocess import CalledProcessError

from ._common import exec_in_terminal


def get_global_mpi_info():
    info = {}
    try:
        info["mpicc"] = exec_in_terminal(["which", "mpicc"])
    except CalledProcessError:
        info["mpicc"] = ""

    try:
        info["mpirun"] = exec_in_terminal(["which", "mpirun"])
    except CalledProcessError:
        info["mpirun"] = ""

    try:
        info["mpiexec"] = exec_in_terminal(["which", "mpiexec"])
    except CalledProcessError:
        info["mpiexec"] = ""

    if info["mpicc"]:
        info["link_flags"] = get_link_flags(exec_in_terminal([info["mpicc"], "-show"]))
    else:
        info["link_flags"] = ""
    return info


def get_link_flags(flags):
    flags = flags.replace(",", " ").split()
    res = []
    for flag in flags:
        if flag.startswith("-L"):
            res.append(flag)

    return res



# Copyright 2020 The Netket Authors. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket import legacy as nk
from scipy.sparse.linalg import eigsh

# 1D Lattice
g = nk.graph.Hypercube(length=16, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)

# Heisenberg spin hamiltonian
ha = nk.operator.Heisenberg(hilbert=hi)

# Convert hamiltonian to a sparse matrix
# Here we further take only the real part since the Heisenberg Hamiltonian is real
sp_ha = ha.to_sparse().real

# Use scipy sparse diagonalization
vals, vecs = eigsh(sp_ha, k=2, which="SA")
print("eigenvalues with scipy sparse:", vals)

# Explicitely compute energy of ground state
# Doing full dot product
psi = vecs[:, 0]
print("\ng.s. energy:", psi @ sp_ha @ psi)

# Compute energy of first excited state
psi = vecs[:, 1]
print("\nfirst excited energy:", psi @ sp_ha @ psi)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from collections import namedtuple

from netket.utils.numbers import is_scalar

import jax

from ..qgt import QGTAuto
from ..preconditioner import LinearPreconditioner

Preconditioner = namedtuple("Preconditioner", ["object", "solver"])

default_iterative = "cg"
# default_direct = "eigen"


def build_SR(*args, solver_restart: bool = False, **kwargs):
    """
    Construct the structure holding the parameters for using the
    Stochastic Reconfiguration/Natural gradient method.

    Depending on the arguments, an implementation is chosen. For
    details on all possible kwargs check the specific SR implementations
    in the documentation.

    You can also construct one of those structures directly.

    Args:
        diag_shift: Diagonal shift added to the S matrix
        method: (cg, gmres) The specific method.
        iterative: Whever to use an iterative method or not.
        jacobian: Differentiation mode to precompute gradients
                  can be "holomorphic", "R2R", "R2C",
                  None (if they shouldn't be precomputed)
        rescale_shift: Whether to rescale the diagonal offsets in SR according
                       to diagonal entries (only with precomputed gradients)

    Returns:
        The SR parameter structure.
    """

    #  try to understand if this is the old API or new
    # API

    old_api = False
    # new_api = False

    if "matrix" in kwargs:
        # new syntax
        return _SR(*args, **kwargs)

    legacy_kwargs = ["iterative", "method"]
    legacy_solver_kwargs = ["tol", "atol", "maxiter", "M", "restart", "solve_method"]
    for key in legacy_kwargs + legacy_solver_kwargs:
        if key in kwargs:
            old_api = True
            break

    if len(args) > 0:
        if is_scalar(args[0]):  # it's diag_shift
            old_api = True
        # else:
        #    new_api = True

        if len(args) > 1:
            if isinstance(args[1], str):
                old_api = True
        #     else:
        #        new_api = True

    if old_api:
        for (i, arg) in enumerate(args):
            if i == 0:
                # diag shift
                kwargs["diag_shift"] = arg
            elif i == 1:
                kwargs["method"] = arg
            else:
                raise TypeError(
                    "SR takes at most 2 positional arguments but len(args) where provided"
                )

        args = tuple()

        solver = None
        if "iterative" in kwargs:
            kwargs.pop("iterative")
        if "method" in kwargs:
            legacy_solvers = {
                "cg": jax.scipy.sparse.linalg.cg,
                "gmres": jax.scipy.sparse.linalg.gmres,
            }
            if kwargs["method"] not in legacy_solvers:
                raise ValueError(
                    "The old API only supports cg and gmres solvers. "
                    "Migrate to the new API and use any solver from"
                    "jax.scipy.sparse.linalg"
                )
            solver = legacy_solvers[kwargs["method"]]
            kwargs.pop("method")
        else:
            solver = jax.scipy.sparse.linalg.cg

        solver_keys = {}
        has_solver_kw = False
        for key in legacy_solver_kwargs:
            if key in kwargs:
                solver_keys[key] = kwargs[key]
                has_solver_kw = True

        if has_solver_kw:
            solver = partial(solver, **solver_keys)

        kwargs["solver"] = solver

    return _SR(*args, solver_restart=solver_restart, **kwargs)


class SR(LinearPreconditioner):
    pass


# This will become the future implementation once legacy and semi-legacy
# bejaviour is removed
def _SR(
    qgt=None,
    solver=None,
    *,
    diag_shift: float = 0.01,
    solver_restart: bool = False,
    **kwargs,
):
    """
    Construct the structure holding the parameters for using the
    Stochastic Reconfiguration/Natural gradient method.

    Depending on the arguments, an implementation is chosen. For
    details on all possible kwargs check the specific SR implementations
    in the documentation.

    You can also construct one of those structures directly.

    Args:
        qgt: The Quantum Geomtric Tensor type to use.
        diag_shift: Diagonal shift added to the S matrix
        method: (cg, gmres) The specific method.
        iterative: Whever to use an iterative method or not.
        jacobian: Differentiation mode to precompute gradients
                  can be "holomorphic", "R2R", "R2C",
                         None (if they shouldn't be precomputed)
        rescale_shift: Whether to rescale the diagonal offsets in SR according
                       to diagonal entries (only with precomputed gradients)

    Returns:
        The SR parameter structure.
    """

    if solver is None:
        solver = jax.scipy.sparse.linalg.cg

    if qgt is None:
        qgt = QGTAuto(solver)

    return SR(
        partial(qgt, diag_shift=diag_shift, **kwargs),
        solver=solver,
        solver_restart=solver_restart,
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import haiku as hk
import jax.numpy as jnp
import numpy as np


class MyLinear(hk.Module):
    def __init__(self, output_size, name=None):
        super().__init__(name=name)
        self.output_size = output_size

    def __call__(self, x):
        j, k = x.shape[-1], self.output_size
        w_init = hk.initializers.TruncatedNormal(1.0 / np.sqrt(j))
        w = hk.get_parameter("w", shape=[j, k], dtype=x.dtype, init=w_init)
        b = hk.get_parameter("b", shape=[k], dtype=x.dtype, init=jnp.zeros)
        return jnp.dot(x, w) + b


def forward_fn(x):
    model = MyLinear(10)
    return jnp.squeeze(jnp.sum(model(x), axis=-1))


ma = hk.transform(forward_fn)

# 1D Lattice
L = 20  # 10

g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)

# Ising spin hamiltonian
ha = nk.operator.Ising(hilbert=hi, graph=g, h=1.0)

# Metropolis Local Sampling
sa = nk.sampler.MetropolisLocal(hi, n_chains=16)

# Optimizer
op = nk.optimizer.Sgd(learning_rate=0.1)

# Variational monte carlo driver
gs = nk.VMC(ha, op, sa, ma, n_samples=1000, n_discard_per_chain=50)

# Run the optimization for 300 iterations
gs.run(n_iter=300, out="test")



from pytest import approx
import netket as nk
import numpy as np

from .. import common

pytestmark = common.skipif_mpi


def test_ed():
    first_n = 3
    g = nk.graph.Chain(8)
    hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)
    ha = nk.operator.Ising(graph=g, h=1.0, hilbert=hi)

    def expval(op, v):
        return np.vdot(v, op(v))

    # Test Lanczos ED with eigenvectors
    w, v = nk.exact.lanczos_ed(ha, k=first_n, compute_eigenvectors=True)
    assert w.shape == (first_n,)
    assert v.shape == (hi.n_states, first_n)
    gse = expval(ha, v[:, 0])
    fse = expval(ha, v[:, 1])
    assert gse == approx(w[0], rel=1e-14, abs=1e-14)
    assert fse == approx(w[1], rel=1e-14, abs=1e-14)

    # Test Lanczos ED without eigenvectors
    w = nk.exact.lanczos_ed(ha, k=first_n, compute_eigenvectors=False)
    assert w.shape == (first_n,)

    # Test Lanczos ED with custom options
    w_tol = nk.exact.lanczos_ed(
        ha,
        k=first_n,
        scipy_args={"tol": 1e-9, "maxiter": 1000},
    )
    assert w_tol.shape == (first_n,)
    assert w_tol == approx(w)

    # Test Full ED with eigenvectors
    w_full, v_full = nk.exact.full_ed(ha, compute_eigenvectors=True)
    assert w_full.shape == (hi.n_states,)
    assert v_full.shape == (hi.n_states, hi.n_states)
    gse = expval(ha, v_full[:, 0])
    fse = expval(ha, v_full[:, 1])
    assert gse == approx(w_full[0], rel=1e-14, abs=1e-14)
    assert fse == approx(w_full[1], rel=1e-14, abs=1e-14)
    assert w == approx(w_full[:3], rel=1e-14, abs=1e-14)

    # Test Full ED without eigenvectors
    w_full = nk.exact.full_ed(ha, compute_eigenvectors=False)
    assert w_full.shape == (hi.n_states,)
    assert w == approx(w_full[:3], rel=1e-14, abs=1e-14)


def test_ed_restricted():
    g = nk.graph.Hypercube(length=8, n_dim=1, pbc=True)
    hi1 = nk.hilbert.Spin(s=0.5, N=g.n_nodes, total_sz=0)
    hi2 = nk.hilbert.Spin(s=0.5, N=g.n_nodes)

    ham1 = nk.operator.Heisenberg(hi1, graph=g)
    ham2 = nk.operator.Heisenberg(hi2, graph=g)

    assert ham1.to_linear_operator().shape == (70, 70)
    assert ham2.to_linear_operator().shape == (256, 256)

    w1, v1 = nk.exact.lanczos_ed(ham1, compute_eigenvectors=True)
    w2, v2 = nk.exact.lanczos_ed(ham2, compute_eigenvectors=True)

    assert w1[0] == approx(w2[0])

    def overlap(phi, psi):
        bare_overlap = np.abs(np.vdot(phi, psi)) ** 2
        return bare_overlap / (np.vdot(phi, phi) * np.vdot(psi, psi)).real

    # Non-zero elements of ground state in full Hilbert space should equal the ground
    # state in the constrained Hilbert space
    idx_nonzero = np.abs(v2[:, 0]) > 1e-4
    assert overlap(v1[:, 0], v2[:, 0][idx_nonzero]) == approx(1.0)



# Copyright 2021 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.struct."""

from typing import Any
import pytest
from functools import partial

import dataclasses

from netket.utils import struct

import jax

from .. import common

pytestmark = common.skipif_mpi


@struct.dataclass
class Point0:
    x: float
    y: float
    meta: Any = struct.field(pytree_node=False)

    def __pre_init__(self, *args, **kwargs):
        if "z" in kwargs:
            kwargs["x"] = kwargs.pop("z") * 10

        return args, kwargs

    @struct.property_cached
    def cached_node(self) -> int:
        return 3


@struct.dataclass
class Point1:
    x: float
    y: float
    meta: Any = struct.field(pytree_node=False)

    @struct.property_cached
    def cached_node(self) -> int:
        return 3


@struct.dataclass
class Point1Child(Point1):
    z: float

    @struct.property_cached
    def cached_node(self) -> int:
        return 4


@struct.dataclass
class Point1Child2(Point1):
    z: float


Point1ChildConstructor = partial(Point1Child, z=3)
Point1Child2Constructor = partial(Point1Child2, z=3)


@pytest.mark.parametrize(
    "PointT", [Point0, Point1, Point1ChildConstructor, Point1Child2Constructor]
)
def test_no_extra_fields(PointT):
    p = PointT(x=1, y=2, meta={})
    with pytest.raises(dataclasses.FrozenInstanceError):
        p.new_field = 1


@pytest.mark.parametrize(
    "PointT", [Point0, Point1, Point1ChildConstructor, Point1Child2Constructor]
)
def test_mutation(PointT):
    p = PointT(x=1, y=2, meta={})
    new_p = p.replace(x=3)
    assert new_p == PointT(x=3, y=2, meta={})
    with pytest.raises(dataclasses.FrozenInstanceError):
        p.y = 3


@pytest.mark.parametrize("PointT", [Point0, Point1])
def test_pytree_nodes(PointT):
    p = PointT(x=1, y=2, meta={"abc": True})
    leaves = jax.tree_leaves(p)
    assert leaves == [1, 2]
    new_p = jax.tree_map(lambda x: x + x, p)
    assert new_p == PointT(x=2, y=4, meta={"abc": True})


def test_pytree_nodes_inheritance():
    p = Point1Child(x=1, y=2, z=3, meta={"abc": True})
    _ = Point1Child(1, 2, {"abc": True}, 3)
    leaves = jax.tree_leaves(p)
    assert leaves == [1, 2, 3]
    new_p = jax.tree_map(lambda x: x + x, p)
    assert new_p == Point1Child(x=2, y=4, z=6, meta={"abc": True})


@pytest.mark.parametrize("PointT", [Point0, Point1, Point1Child2Constructor])
def test_cached_property(PointT):
    p = PointT(x=1, y=2, meta={"abc": True})

    assert p.__cached_node_cache is struct.Uninitialized
    assert p.cached_node == 3
    assert p.__cached_node_cache == 3

    p = p.replace(x=1)
    assert p.__cached_node_cache is struct.Uninitialized
    p._precompute_cached_properties()
    assert p.__cached_node_cache == 3


def test_cached_property_inheritance():
    p = Point1Child(x=1, y=2, z=3, meta={"abc": True})

    assert p.__cached_node_cache is struct.Uninitialized
    assert p.cached_node == 4
    assert p.__cached_node_cache == 4

    p = p.replace(x=1)
    assert p.__cached_node_cache is struct.Uninitialized
    p._precompute_cached_properties()
    assert p.__cached_node_cache == 4


def test_pre_init_property():
    p = Point0(z=1, y=2, meta={"abc": True})

    assert p.x == 10


def test_inheritance():
    p = Point1Child(x=1, z=1, y=2, meta={"abc": True})

    assert p.x == 1
    assert p.z == 1


@struct.dataclass(cache_hash=True)
class Point0cache:
    x: float

    def __hash__(self):
        return 123


def test_cache_hash():
    a = Point0cache(1)
    assert a.__Point0cache_hash_cache is struct.Uninitialized
    hash(a) == 123
    assert a.__Point0cache_hash_cache == 123
    object.__setattr__(a, "__Point0cache_hash_cache", 1234)
    hash(a) == 1234


@struct.dataclass
class PointC:
    x: float
    y: float

    @struct.property_cached(pytree_node=True)
    def cached_node(self) -> int:
        return self.x * self.y


# test for pytree_node=True cached properties
def test_cached_pytreenode_properties():
    # check cache uninitialized then populated
    p = PointC(2.0, 3.0)
    assert p.__cached_node_cache is struct.Uninitialized
    assert p.cached_node == 2.0 * 3.0
    assert p.__cached_node_cache == 6.0

    # check cache reset after call to replace
    p = p.replace(y=3.0)
    assert p.__cached_node_cache is struct.Uninitialized
    assert p.cached_node == 2.0 * 3.0
    assert p.__cached_node_cache == 6.0

    @jax.jit
    def compute(x: PointC):
        return x.cached_node * 3, x

    # check cache populated when returned from jitted function
    p = PointC(2.0, 3.0)
    res, p2 = compute(p)
    assert p.__cached_node_cache is struct.Uninitialized
    assert res == 6.0 * 3
    assert p2.__cached_node_cache == 6.0

    # check cache not modified by jit
    res, p3 = compute(p2)
    assert res == 6.0 * 3
    assert p3.__cached_node_cache == 6.0



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import numpy as np
import jax.numpy as jnp

import jax

from .test_nn import _setup_symm

import pytest


@pytest.mark.parametrize("use_hidden_bias", [True, False])
@pytest.mark.parametrize("use_visible_bias", [True, False])
@pytest.mark.parametrize("symmetries", ["trans", "autom"])
def test_RBMSymm(use_hidden_bias, use_visible_bias, symmetries):
    g, hi, perms = _setup_symm(symmetries, N=8)

    ma = nk.models.RBMSymm(
        symmetries=perms,
        alpha=4,
        use_visible_bias=use_visible_bias,
        use_hidden_bias=use_hidden_bias,
        hidden_bias_init=nk.nn.initializers.uniform(),
        visible_bias_init=nk.nn.initializers.uniform(),
    )
    pars = ma.init(nk.jax.PRNGKey(), hi.random_state(nk.jax.PRNGKey()))

    print(pars)

    v = hi.random_state(jax.random.PRNGKey(1), 3)
    vals = [ma.apply(pars, v[..., p]) for p in np.asarray(perms)]

    for val in vals:
        assert jnp.allclose(val, vals[0])

    vmc = nk.VMC(
        nk.operator.Ising(hi, g, h=1.0),
        nk.optim.Sgd(0.1),
        nk.sampler.MetropolisLocal(hi),
        ma,
    )
    vmc.advance(1)


@pytest.mark.parametrize("parity", [True, False])
@pytest.mark.parametrize("symmetries", ["trans", "autom"])
@pytest.mark.parametrize("lattice", [nk.graph.Chain, nk.graph.Square])
@pytest.mark.parametrize("mode", ["fft", "irreps"])
def test_gcnn(parity, symmetries, lattice, mode):
    g, hi, perms = _setup_symm(symmetries, N=3, lattice=lattice)

    ma = nk.models.GCNN(
        symmetries=perms,
        mode=mode,
        shape=tuple(g.extent),
        layers=2,
        features=2,
        parity=parity,
        bias_init=nk.nn.initializers.uniform(),
    )

    pars = ma.init(nk.jax.PRNGKey(), hi.random_state(nk.jax.PRNGKey(), 1))

    v = hi.random_state(jax.random.PRNGKey(0), 3)
    vals = [ma.apply(pars, v[..., p]) for p in np.asarray(perms)]

    for val in vals:
        assert jnp.allclose(val, vals[0])

    vmc = nk.VMC(
        nk.operator.Ising(hi, g, h=1.0),
        nk.optim.Sgd(0.1),
        nk.sampler.MetropolisLocal(hi),
        ma,
    )
    vmc.advance(1)


def test_RBMSymm_creation():
    hi = nk.hilbert.Spin(1 / 2, N=8)

    def check_init(creator):
        ma = creator()
        _ = ma.init(nk.jax.PRNGKey(0), hi.numbers_to_states(0))

    perms = [[0, 1, 2, 3, 4, 5, 6, 7]]

    # Test different permutation argument types
    check_init(lambda: nk.models.RBMSymm(symmetries=perms))
    check_init(lambda: nk.models.RBMSymm(symmetries=jnp.array(perms)))

    # wrong shape
    with pytest.raises(ValueError):
        check_init(lambda: nk.models.RBMSymm(symmetries=perms[0]))

    # init with PermutationGroup
    check_init(
        lambda: nk.models.RBMSymm(
            symmetries=nk.graph.Chain(8).translation_group(), alpha=2
        )
    )

    # alpha too small
    with pytest.raises(ValueError):
        check_init(
            lambda: nk.models.RBMSymm(
                symmetries=nk.graph.Hypercube(8, 2).automorphisms(), alpha=1
            )
        )


@pytest.mark.parametrize("mode", ["fft", "irreps"])
def test_GCNN_creation(mode):

    g = nk.graph.Chain(8)
    space_group = g.space_group()
    hi = nk.hilbert.Spin(1 / 2, N=8)

    def check_init(creator):
        ma = creator()
        _ = ma.init(nk.jax.PRNGKey(0), hi.numbers_to_states(0))

    perms = [[0, 1, 2, 3, 4, 5, 6, 7], [7, 6, 5, 4, 3, 2, 1, 0]]

    # Init with graph
    check_init(
        lambda: nk.models.GCNN(
            symmetries=g,
            mode=mode,
            layers=2,
            features=4,
        )
    )

    # init with space_group
    if mode == "irreps":
        check_init(
            lambda: nk.models.GCNN(
                symmetries=space_group,
                mode=mode,
                layers=2,
                features=4,
            )
        )
    else:
        check_init(
            lambda: nk.models.GCNN(
                symmetries=space_group,
                shape=tuple(g.extent),
                mode=mode,
                layers=2,
                features=4,
            )
        )

    # init with arrays for sym and product_table
    if mode == "irreps":
        check_init(
            lambda: nk.models.GCNN(
                symmetries=np.asarray(space_group),
                irreps=space_group.irrep_matrices(),
                layers=2,
                features=4,
            )
        )
    else:
        check_init(
            lambda: nk.models.GCNN(
                symmetries=np.asarray(space_group),
                product_table=space_group.product_table,
                shape=(8,),
                layers=2,
                features=4,
            )
        )

    # forget irreps/product table
    with pytest.raises(ValueError):
        check_init(
            lambda: nk.models.GCNN(
                symmetries=perms[0],
                layers=2,
                features=4,
            )
        )

    # need to specify shape
    if mode == "fft":
        with pytest.raises(TypeError):
            check_init(
                lambda: nk.models.GCNN(
                    symmetries=perms,
                    product_table=np.arange(4).reshape(2, 2),
                    layers=2,
                    features=4,
                )
            )

    # character table
    check_init(
        lambda: nk.models.GCNN(
            symmetries=g,
            mode=mode,
            layers=2,
            features=4,
            characters=np.ones([len(np.asarray(space_group))]),
        )
    )

    # equal amplitudes
    check_init(
        lambda: nk.models.GCNN(
            symmetries=g, mode=mode, layers=2, features=4, equal_amplitudes=True
        )
    )


@pytest.mark.parametrize("use_hidden_bias", [True, False])
@pytest.mark.parametrize("use_visible_bias", [True, False])
def test_RBMMultiVal(use_hidden_bias, use_visible_bias):
    N = 8
    M = 3
    hi = nk.hilbert.Fock(M, N)
    g = nk.graph.Chain(N)

    ma = nk.models.RBMMultiVal(
        alpha=2,
        n_classes=M + 1,
        use_visible_bias=use_visible_bias,
        use_hidden_bias=use_hidden_bias,
        hidden_bias_init=nk.nn.initializers.uniform(),
        visible_bias_init=nk.nn.initializers.uniform(),
    )
    _ = ma.init(nk.jax.PRNGKey(), hi.random_state(nk.jax.PRNGKey(), 1))

    vmc = nk.VMC(
        nk.operator.BoseHubbard(hi, g, U=1.0),
        nk.optim.Sgd(0.1),
        nk.sampler.MetropolisLocal(hi),
        ma,
    )
    vmc.advance(1)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

import netket as nk

from .. import common


@common.onlyif_mpi
def test_key_split(_mpi_size, _mpi_comm):
    from mpi4py import MPI

    size = _mpi_size
    comm = _mpi_comm

    key = nk.jax.PRNGKey(1256)

    keys = comm.allgather(key)
    assert all([jnp.all(k == key) for k in keys])

    key, _ = nk.jax.mpi_split(key)
    keys = MPI.COMM_WORLD.allgather(key)
    assert all([not jnp.all(k == keys) for k in keys])
    assert len(keys) == size



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from plum import dispatch, parametric, convert  # noqa: F401


# A simple type to represent a compile-constant True/False type
class Bool:
    """A class representing a static True/False that can be used for dispatch."""

    pass


class TrueT(Bool):
    """A class representing a static True value that can be used for dispatch."""

    def __bool__(self):
        return True


class FalseT(Bool):
    """A class representing a static False value that can be used for dispatch."""

    def __bool__(self):
        return False



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket
import numpy as _np

__all__ = ["PyRbm"]


class PyRbm(netket.machine.CxxMachine):
    r"""
    __Do not use me in production code!__

    A proof of concept implementation of a complex-valued RBM in pure Python.
    This is an example of how to subclass `CxxMachine` so that the machine will
    be usable with NetKet's C++ core.

    This class can be used as a drop-in replacement for `RbmSpin`.
    """

    def __init__(
        self, hilbert, alpha=None, use_visible_bias=True, use_hidden_bias=True
    ):
        r"""Constructs a new RBM.

        Args:
            hilbert: Hilbert space.
            alpha: `alpha * hilbert.size` is the number of hidden spins.
            use_visible_bias: specifies whether to use a bias for visible
                              spins.
            use_hidden_bias: specifies whether to use a bias for hidden spins.
        """
        # NOTE: The following call to __init__ is important!
        super(PyRbm, self).__init__(hilbert, dtype=complex)
        n = hilbert.size
        if alpha < 0:
            raise ValueError("`alpha` should be non-negative")
        m = int(round(alpha * n))
        self._w = _np.empty([m, n], dtype=_np.complex128)
        self._a = _np.empty(n, dtype=_np.complex128) if use_visible_bias else None
        self._b = _np.empty(m, dtype=_np.complex128) if use_hidden_bias else None

    def _n_par(self):
        r"""Returns the number of parameters in the machine. We just sum the
        sizes of all the tensors we hold.
        """
        return (
            self._w.size
            + (self._a.size if self._a is not None else 0)
            + (self._b.size if self._b is not None else 0)
        )

    def _log_val(self, x, out):
        r"""Computes the logarithm of the wave function given a spin
        configuration ``x``.
        """
        r = _np.dot(x, self._w.T)
        if self._b is not None:
            r += self._b

        _np.sum(PyRbm._log_cosh(r), axis=-1, out=out)

        if self._a is not None:
            out += _np.dot(x, self._a)
        # Officially, we should return
        #     self._w.shape[0] * 0.6931471805599453 + r
        # but the C++ implementation ignores the "constant factor"

    def _der_log(self, x, out):

        batch_size = x.shape[0]

        i = 0
        if self._a is not None:
            out[:, i : i + x.shape[1]] = x
            i += x.shape[1]

        r = _np.dot(x, self._w.T)
        if self._b is not None:
            r += self._b
        _np.tanh(r, out=r)

        if self._b is not None:
            out[:, i : i + self._b.shape[0]] = r
            i += self._b.shape[0]

        t = out[:, i : i + self._w.size]
        t.shape = (batch_size, self._w.shape[0], self._w.shape[1])
        _np.einsum("ij,il->ijl", r, x, out=t)

    def state_dict(self):
        from collections import OrderedDict

        return OrderedDict(
            [("a", self._a.view()), ("b", self._b.view()), ("w", self._w.view())]
        )

    @staticmethod
    def _log_cosh(x):
        # TODO: Handle big numbers properly
        return _np.log(_np.cosh(x))



import jax
from jax import numpy as jnp

from netket.hilbert import AbstractParticle
from netket.utils.dispatch import dispatch


@dispatch
def random_state(hilb: AbstractParticle, key, batches: int, *, dtype):
    """Positions particles w.r.t. normal distribution,
    if no periodic boundary conditions are applied
    in a spatial dimension. Otherwise the particles are
    positioned evenly along the box from 0 to L, with Gaussian noise
    of certain width."""
    pbc = jnp.array(hilb.n_particles * hilb.pbc)
    boundary = jnp.tile(pbc, (batches, 1))

    Ls = jnp.array(hilb.n_particles * hilb.extend)
    modulus = jnp.where(jnp.equal(pbc, False), jnp.inf, Ls)

    gaussian = jax.random.normal(key, shape=(batches, hilb.size))
    width = jnp.min(modulus) / (4.0 * hilb.n_particles)
    # The width gives the noise level. In the periodic case the
    # particles are evenly distributed between 0 and min(L). The
    # distance between the particles coordinates is therefore given by
    # min(L) / hilb.N. To avoid particles to have coincident
    # positions the noise level should be smaller than half this distance.
    # We choose width = min(L) / (4*hilb.N)
    noise = gaussian * width
    uniform = jnp.tile(jnp.linspace(0.0, jnp.min(modulus), hilb.size), (batches, 1))

    rs = jnp.where(jnp.equal(boundary, False), gaussian, (uniform + noise) % modulus)

    return jnp.asarray(rs, dtype=dtype)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import common


@common.onlyif_mpi
def test_mpi_setup(_mpi_rank, _mpi_size, _mpi_comm):
    rank = _mpi_rank
    size = _mpi_size
    comm = _mpi_comm

    recv = comm.bcast(rank)
    assert recv == 0, "rank={}".format(rank)

    ranks = comm.allgather(rank)
    assert len(ranks) == size
    assert ranks == list(range(size))



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
from typing import Union, List, Optional
from netket.utils.types import DType, Array
from textwrap import dedent

import numpy as np
from numba import jit

from netket.hilbert import AbstractHilbert, Fock

from ._abstract_operator import AbstractOperator
from ._lazy import Transpose


@jit(nopython=True)
def _number_to_state(number, hilbert_size_per_site, local_states_per_site, out):

    out[:] = local_states_per_site[:, 0]
    size = out.shape[0]

    ip = number
    k = size - 1
    while ip > 0:
        local_size = hilbert_size_per_site[k]
        out[k] = local_states_per_site[k, ip % local_size]
        ip = ip // local_size
        k -= 1

    return out


def is_hermitian(a: np.ndarray, rtol=1e-05, atol=1e-08) -> bool:
    return np.allclose(a, a.T.conj(), rtol=rtol, atol=atol)


def _dtype(obj: Union[numbers.Number, Array, "LocalOperator"]) -> DType:
    if isinstance(obj, numbers.Number):
        return type(obj)
    elif isinstance(obj, AbstractOperator):
        return obj.dtype
    elif isinstance(obj, np.ndarray):
        return obj.dtype
    else:
        raise TypeError(f"cannot deduce dtype of object type {type(obj)}: {obj}")


def has_nonzero_diagonal(op: "LocalOperator") -> bool:
    """
    Returns True if at least one element in the diagonal of the operator
    is nonzero.
    """
    return (
        np.any(np.abs(op._diag_mels) >= op.mel_cutoff)
        or np.abs(op._constant) >= op.mel_cutoff
    )


def _is_sorted(a):
    for i in range(len(a) - 1):
        if a[i + 1] < a[i]:
            return False
    return True


def resize(
    arr: Array,
    shape: List[int],
    dtype: Optional[DType] = None,
    init: Optional[numbers.Number] = None,
) -> Array:
    """
    resizes the input array to the new shape that must be larger than the old.

    The resulting array is initialized with the old array in the corresponding indices, and with init
    in the rest.

    Args:
        arr: The array to be resized
        shape: The new shape
        dtype: optional dtype of the new array. If unspecified the old array dtype is used
        init: optional initialization value for the new entries

    Returns:
        a numpy array with the chosen shape and dtype.
    """
    if dtype is None:
        dtype = arr.dtype

    if isinstance(shape, int):
        shape = (shape,)

    if arr.shape == shape:
        return arr

    arr_shape = arr.shape
    if len(shape) != arr.ndim:
        raise ValueError("the number of dimensions should not change.")

    for (i, ip) in zip(arr_shape, shape):
        if ip < i:
            raise ValueError(
                f"The new dimensions ({shape}) should all be larger than the old ({arr_shape})."
            )

    new_arr = np.empty(shape=shape, dtype=arr.dtype)
    if init is not None:
        new_arr[...] = init

    if arr.ndim == 0:
        raise ValueError("Cannot resize a 0-dimensional scalar")
    elif arr.ndim == 1:
        new_arr[: arr_shape[0]] = arr
    elif arr.ndim == 2:
        new_arr[: arr_shape[0], : arr_shape[1]] = arr
    elif arr.ndim == 3:
        new_arr[: arr_shape[0], : arr_shape[1], : arr_shape[2]] = arr
    elif arr.ndim == 4:
        new_arr[: arr_shape[0], : arr_shape[1], : arr_shape[2], : arr_shape[3]] = arr
    else:
        raise ValueError(f"unsupported number of dimensions: {arr.ndim}")

    return new_arr


def _reorder_kronecker_product(hi, mat, acting_on):
    """
    Reorders the matrix resulting from a kronecker product of several
    operators in such a way to sort acting_on.

    A conceptual example is the following:
    if `mat = Â ⊗ B̂ ⊗ Ĉ` and `acting_on = [[2],[1],[3]`
    you will get `result = B̂ ⊗ Â ⊗ Ĉ, [[1], [2], [3]].

    However, essentially, A,B,C represent some operators acting on
    thei sub-space acting_on[1], [2] and [3] of the hilbert space.

    This function also handles any possible set of values in acting_on.

    The inner logic uses the Fock.all_states(), number_to_state and
    state_to_number to perform the re-ordering.
    """
    acting_on_sorted = np.sort(acting_on)
    if np.all(acting_on_sorted == acting_on):
        return mat, acting_on

    # could write custom binary <-> int logic instead of using Fock...
    # Since i need to work with bit-strings (where instead of bits i
    # have integers, in order to support arbitrary size spaces) this
    # is exactly what hilbert.to_number() and viceversa do.

    # target ordering binary representation
    hi_subspace = Fock(hi.shape[acting_on_sorted[0]] - 1)
    for site in acting_on_sorted[1:]:
        hi_subspace = Fock(hi.shape[site] - 1) * hi_subspace

    # find how to map target ordering back to unordered
    acting_on_unsorted_ids = np.zeros(len(acting_on), dtype=np.intp)
    for (i, site) in enumerate(acting_on):
        acting_on_unsorted_ids[i] = np.argmax(site == acting_on_sorted)

    # now it is valid that
    # acting_on_sorted == acting_on[acting_on_unsorted_ids]

    # generate n-bit strings in the target ordering
    v = hi_subspace.all_states()

    # convert them to origin (unordered) ordering
    v_unsorted = v[:, acting_on_unsorted_ids]
    # convert the unordered bit-strings to numbers in the target space.
    n_unsorted = hi_subspace.states_to_numbers(v_unsorted)

    # reorder the matrix
    mat_sorted = mat[n_unsorted, :][:, n_unsorted]

    return mat_sorted, acting_on_sorted


class LocalOperator(AbstractOperator):
    """A custom local operator. This is a sum of an arbitrary number of operators
    acting locally on a limited set of k quantum numbers (i.e. k-local,
    in the quantum information sense).
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        operators: Union[List[Array], Array] = [],
        acting_on: Union[List[int], List[List[int]]] = [],
        constant: numbers.Number = 0,
        dtype: Optional[DType] = None,
    ):
        r"""
        Constructs a new ``LocalOperator`` given a hilbert space and (if
        specified) a constant level shift.

        Args:
           hilbert (netket.AbstractHilbert): Hilbert space the operator acts on.
           operators (list(numpy.array) or numpy.array): A list of operators, in matrix form.
           acting_on (list(numpy.array) or numpy.array): A list of sites, which the corresponding operators act on.
           constant (float): Level shift for operator. Default is 0.0.

        Examples:
           Constructs a ``LocalOperator`` without any operators.

           >>> from netket.hilbert import CustomHilbert
           >>> from netket.operator import LocalOperator
           >>> hi = CustomHilbert(local_states=[-1, 1])**20
           >>> empty_hat = LocalOperator(hi)
           >>> print(len(empty_hat.acting_on))
           0
        """
        super().__init__(hilbert)
        self._constant = constant

        if not all(
            [_is_sorted(hilbert.states_at_index(i)) for i in range(hilbert.size)]
        ):
            raise ValueError(
                dedent(
                    """LocalOperator needs an hilbert space with sorted state values at
                every site.
                """
                )
            )

        # check if passing a single operator or a list of operators
        if isinstance(acting_on, numbers.Number):
            acting_on = [acting_on]

        is_nested = any(hasattr(i, "__len__") for i in acting_on)

        if not is_nested:
            operators = [operators]
            acting_on = [acting_on]

        operators = [np.asarray(operator) for operator in operators]

        # If we asked for a specific dtype, enforce it.
        if dtype is None:
            dtype = np.promote_types(operators[0].dtype, np.float32)
            for op in operators[1:]:
                np.promote_types(dtype, op.dtype)

        self._dtype = dtype
        self._init_zero()

        self.mel_cutoff = 1.0e-6

        self._nonzero_diagonal = np.abs(self._constant) >= self.mel_cutoff
        """True if at least one element in the diagonal of the operator is
        nonzero"""

        for op, act in zip(operators, acting_on):
            if len(act) > 0:
                self._add_operator(op, act)

    @property
    def operators(self) -> List[np.ndarray]:
        """List of the matrices of the operators encoded in this Local Operator.
        Returns a copy.
        """
        return self._operators_list()

    @property
    def acting_on(self) -> List[List[int]]:
        """List containing the list of the sites on which every operator acts.

        Every operator `self.operators[i]` acts on the sites `self.acting_on[i]`
        """
        actions = [action[action >= 0].tolist() for action in self._acting_on]
        return actions

    @property
    def dtype(self) -> DType:
        return self._dtype

    @property
    def size(self) -> int:
        return self._size

    @property
    def is_hermitian(self) -> bool:
        """Returns true if this operator is hermitian."""
        return self._is_hermitian

    @property
    def mel_cutoff(self) -> float:
        r"""float: The cutoff for matrix elements.
        Only matrix elements such that abs(O(i,i))>mel_cutoff
        are considered"""
        return self._mel_cutoff

    @mel_cutoff.setter
    def mel_cutoff(self, mel_cutoff):
        self._mel_cutoff = mel_cutoff
        assert self.mel_cutoff >= 0

    @property
    def constant(self) -> numbers.Number:
        return self._constant

    @property
    def n_operators(self) -> int:
        return self._n_operators

    def __add__(self, other: Union["LocalOperator", numbers.Number]):
        op = self.copy(dtype=np.promote_types(self.dtype, _dtype(other)))
        op = op.__iadd__(other)
        return op

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        if isinstance(other, LocalOperator):
            if self.hilbert != other.hilbert:
                return NotImplemented

            if not np.can_cast(other.dtype, self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {other.dtype} to operator with dtype {self.dtype}"
                )

            assert other.mel_cutoff == self.mel_cutoff

            for i in range(other._n_operators):
                acting_on = other._acting_on[i, : other._acting_size[i]]
                operator = other._operators[i]
                self._add_operator(operator, acting_on)

            self._constant += other.constant
            self._nonzero_diagonal = has_nonzero_diagonal(self)

            return self
        if isinstance(other, numbers.Number):

            if not np.can_cast(type(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {type(other)} to operator with dtype {self.dtype}"
                )

            self._constant += other
            self._nonzero_diagonal = has_nonzero_diagonal(self)
            return self

        return NotImplemented

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __isub__(self, other):
        return self.__iadd__(-other)

    def __neg__(self):
        return -1 * self

    def __mul__(self, other):
        if isinstance(other, AbstractOperator):
            op = self.copy(dtype=np.promote_types(self.dtype, _dtype(other)))
            return op.__imatmul__(other)
        elif not isinstance(other, numbers.Number):
            return NotImplemented

        op = self.copy(dtype=np.promote_types(self.dtype, _dtype(other)))

        op._diag_mels *= other
        op._mels *= other
        op._constant *= other

        for _op in op._operators:
            _op *= other

        op._nonzero_diagonal = has_nonzero_diagonal(op)

        return op

    def __imul__(self, other):
        if isinstance(other, AbstractOperator):
            return self.__imatmul__(other)
        elif not isinstance(other, numbers.Number):
            return NotImplemented

        if not np.can_cast(type(other), self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot add inplace operator with dtype {type(other)} to operator with dtype {self.dtype}"
            )

        self._diag_mels *= other
        self._mels *= other
        self._constant *= other

        for _op in self._operators:
            _op *= other

        self._nonzero_diagonal = has_nonzero_diagonal(self)

        return self

    def __imatmul__(self, other):
        if not isinstance(other, LocalOperator):
            return NotImplemented

        if not np.can_cast(other.dtype, self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot add inplace operator with dtype {type(other)} to operator with dtype {self.dtype}"
            )

        return self._concrete_imatmul_(other)

    def _op__matmul__(self, other):
        return self._concrete_matmul_(other)

    def _concrete_matmul_(self, other: "LocalOperator") -> "LocalOperator":
        if not isinstance(other, LocalOperator):
            return NotImplemented
        op = self.copy(dtype=np.promote_types(self.dtype, _dtype(other)))
        op @= other
        return op

    def _concrete_imatmul_(self, other: "LocalOperator") -> "LocalOperator":
        if not isinstance(other, LocalOperator):
            return NotImplemented

        tot_operators = []
        tot_act = []
        for i in range(other._n_operators):
            act_i = other._acting_on[i, : other._acting_size[i]].tolist()
            ops, act = self._multiply_operator(other._operators[i], act_i)
            tot_operators += ops
            tot_act += act

        prod = LocalOperator(self.hilbert, tot_operators, tot_act, dtype=self.dtype)
        self_constant = self._constant
        if np.abs(other._constant) > self.mel_cutoff:
            self *= other._constant
            self += prod
            self._constant = 0.0
        else:
            self = prod

        if np.abs(self_constant) > self.mel_cutoff:
            self += other * self_constant

        self._nonzero_diagonal = has_nonzero_diagonal(self)

        return self

    def __truediv__(self, other):
        if not isinstance(other, numbers.Number):
            raise TypeError("Only divison by a scalar number is supported.")

        if other == 0:
            raise ValueError("Dividing by 0")
        return self.__mul__(1.0 / other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def _init_zero(self):
        self._operators = []
        self._n_operators = 0

        self._max_op_size = 0
        self._max_acting_size = 0
        self._max_local_hilbert_size = 0
        self._size = 0

        self._acting_on = np.zeros((0, 0), dtype=np.intp)
        self._acting_size = np.zeros(0, dtype=np.intp)
        self._diag_mels = np.zeros((0, 0), dtype=self.dtype)
        self._mels = np.empty((0, 0, 0), dtype=self.dtype)
        self._x_prime = np.empty((0, 0, 0, 0))
        self._n_conns = np.empty((0, 0), dtype=np.intp)

        self._local_states = np.zeros((0, 0, 0), dtype=np.float64)

        self._basis = np.zeros((0, 0), dtype=np.int64)
        self._is_hermitian = True

    def _acting_on_list(self):
        acting_on = []
        for i in range(self.n_operators):
            acting_on.append(np.copy(self._acting_on[i, : self._acting_size[i]]))

        return acting_on

    def _operators_list(self):
        "A deep copy of the operators"
        operators = [np.copy(op) for op in self._operators]
        return operators

    def _add_operator(self, operator: Array, acting_on: List[int]):
        if not np.can_cast(operator, self.dtype, casting="same_kind"):
            raise ValueError(f"Cannot cast type {operator.dtype} to {self.dtype}")

        acting_on = np.asarray(acting_on, dtype=np.intp)
        operator = np.asarray(operator, dtype=self.dtype)

        if np.unique(acting_on).size != acting_on.size:
            raise ValueError("acting_on contains repeated entries.")

        if any(acting_on >= self.hilbert.size):
            raise ValueError("acting_on points to a site not in the hilbert space.")

        if operator.ndim != 2:
            raise ValueError("The operator should be a matrix")

        if np.all(np.abs(operator) < self.mel_cutoff):
            return

        # re-sort the operator
        operator, acting_on = _reorder_kronecker_product(
            self.hilbert, operator, acting_on
        )

        # find overlapping support
        support_i = None
        for (i, support) in enumerate(self._acting_on_list()):
            if np.all(acting_on == support):
                support_i = i
                break

        # If overlapping support, add the local operators themselves
        if support_i is not None:
            dim = min(operator.shape[0], self._operators[support_i].shape[0])
            _opv = self._operators[support_i][:dim, :dim]
            _opv += operator[:dim, :dim]

            n_local_states_per_site = np.asarray(
                [self.hilbert.size_at_index(i) for i in acting_on]
            )

            self._append_matrix(
                self._operators[support_i],
                self._diag_mels[support_i],
                self._mels[support_i],
                self._x_prime[support_i],
                self._n_conns[support_i],
                self._acting_size[support_i],
                self._local_states[support_i],
                self.mel_cutoff,
                n_local_states_per_site,
            )

            isherm = True
            for op in self._operators:
                isherm = isherm and is_hermitian(op)

            self._is_hermitian = isherm
            self._nonzero_diagonal = has_nonzero_diagonal(self)
        else:
            self.__add_new_operator__(operator, acting_on)

    def __add_new_operator__(self, operator: np.ndarray, acting_on: np.ndarray):
        # Else, we must add a completely new operator
        self._n_operators += 1
        self._operators.append(operator)

        # Add a new row and eventually resize the acting_on
        self._acting_size = np.resize(self._acting_size, (self.n_operators,))
        self._acting_size[-1] = acting_on.size
        acting_size = acting_on.size

        self._max_op_size = max((operator.shape[0], self._max_op_size))

        n_local_states_per_site = np.asarray(
            [self.hilbert.size_at_index(i) for i in acting_on]
        )

        if operator.shape[0] != np.prod(n_local_states_per_site):
            raise RuntimeError(
                r"""the given operator matrix has shape={} and acts on
                    the sites={}, which have a local hilbert space size of
                    sizes={} giving an expected shape
                    for the operator expected_shape={}.""".format(
                    operator.shape,
                    acting_on,
                    n_local_states_per_site,
                    np.prod(n_local_states_per_site),
                )
            )

        self._max_acting_size = max(self._max_acting_size, acting_on.size)
        self._max_local_hilbert_size = max(
            self._max_local_hilbert_size, np.max(n_local_states_per_site)
        )

        self._acting_on = resize(
            self._acting_on, shape=(self.n_operators, self._max_acting_size), init=-1
        )
        self._acting_on[-1, :acting_size] = acting_on
        if (
            self._acting_on[-1, :acting_size].max() > self.hilbert.size
            or self._acting_on[-1, :acting_size].min() < 0
        ):
            raise ValueError("Operator acts on an invalid set of sites")

        self._local_states = resize(
            self._local_states,
            shape=(
                self.n_operators,
                self._max_acting_size,
                self._max_local_hilbert_size,
            ),
            init=np.nan,
        )
        ## add an operator to local_states
        for site in range(acting_size):
            self._local_states[-1, site, : n_local_states_per_site[site]] = np.asarray(
                self.hilbert.states_at_index(acting_on[site])
            )
        ## add an operator to basis
        self._basis = resize(
            self._basis, shape=(self.n_operators, self._max_acting_size), init=1e10
        )
        ba = 1
        for s in range(acting_on.size):
            self._basis[-1, s] = ba
            ba *= n_local_states_per_site[acting_on.size - s - 1]
        ##

        self._diag_mels = resize(
            self._diag_mels, shape=(self.n_operators, self._max_op_size), init=np.nan
        )
        self._mels = resize(
            self._mels,
            shape=(self.n_operators, self._max_op_size, self._max_op_size - 1),
            init=np.nan,
        )
        self._x_prime = resize(
            self._x_prime,
            shape=(
                self.n_operators,
                self._max_op_size,
                self._max_op_size - 1,
                self._max_acting_size,
            ),
            init=-1,
        )
        self._n_conns = resize(
            self._n_conns, shape=(self.n_operators, self._max_op_size), init=-1
        )
        if acting_on.max() + 1 >= self._size:
            self._size = acting_on.max() + 1

        self._append_matrix(
            operator,
            self._diag_mels[-1],
            self._mels[-1],
            self._x_prime[-1],
            self._n_conns[-1],
            self._acting_size[-1],
            self._local_states[-1],
            self.mel_cutoff,
            n_local_states_per_site,
        )

        isherm = True
        for op in self._operators:
            isherm = isherm and is_hermitian(op)

        self._is_hermitian = isherm

        self._nonzero_diagonal = has_nonzero_diagonal(self)

    @staticmethod
    @jit(nopython=True)
    def _append_matrix(
        operator,
        diag_mels,
        mels,
        x_prime,
        n_conns,
        acting_size,
        local_states_per_site,
        epsilon,
        hilb_size_per_site,
    ):
        op_size = operator.shape[0]
        assert op_size == operator.shape[1]
        for i in range(op_size):
            diag_mels[i] = operator[i, i]
            n_conns[i] = 0
            for j in range(op_size):
                if i != j and np.abs(operator[i, j]) > epsilon:
                    k_conn = n_conns[i]
                    mels[i, k_conn] = operator[i, j]
                    _number_to_state(
                        j,
                        hilb_size_per_site,
                        local_states_per_site[:acting_size, :],
                        x_prime[i, k_conn, :acting_size],
                    )
                    n_conns[i] += 1

    def _multiply_operator(self, op, act):
        operators = []
        acting_on = []
        act = np.asarray(act)

        for i in range(self.n_operators):
            act_i = self._acting_on[i, : self._acting_size[i]]

            inters = np.intersect1d(act_i, act, return_indices=False)

            if act.size == act_i.size and np.array_equal(act, act_i):
                # non-interesecting with same support
                operators.append(np.copy(np.matmul(self._operators[i], op)))
                acting_on.append(act_i.tolist())
            elif inters.size == 0:
                # disjoint supports
                operators.append(np.copy(np.kron(self._operators[i], op)))
                acting_on.append(act_i.tolist() + act.tolist())
            else:
                _act = list(act)
                _act_i = list(act_i)
                _op = op.copy()
                _op_i = self._operators[i].copy()

                # expand _act to match _act_i
                actmin = min(act)
                for site in act_i:
                    if site not in act:
                        I = np.eye(self.hilbert.shape[site], dtype=self.dtype)
                        if site < actmin:
                            _act = [site] + _act
                            _op = np.kron(I, _op)
                        else:  #  site > actmax
                            _act = _act + [site]
                            _op = np.kron(_op, I)

                act_i_min = min(act_i)
                for site in act:
                    if site not in act_i:
                        I = np.eye(self.hilbert.shape[site], dtype=self.dtype)
                        if site < act_i_min:
                            _act_i = [site] + _act_i
                            _op_i = np.kron(I, _op_i)
                        else:  #  site > actmax
                            _act_i = _act_i + [site]
                            _op_i = np.kron(_op_i, I)

                # reorder
                _op, _act = _reorder_kronecker_product(self.hilbert, _op, _act)
                _op_i, _act_i = _reorder_kronecker_product(self.hilbert, _op_i, _act_i)

                if len(_act) == len(_act_i) and np.array_equal(_act, _act_i):
                    # non-interesecting with same support
                    operators.append(np.matmul(_op_i, _op))
                    acting_on.append(_act_i)
                else:
                    raise ValueError("Something failed")

        return operators, acting_on

    def copy(self, *, dtype: Optional = None):
        """Returns a copy of the operator, while optionally changing the dtype
        of the operator.

        Args:
            dtype: optional dtype
        """

        if dtype is None:
            dtype = self.dtype

        if not np.can_cast(self.dtype, dtype, casting="same_kind"):
            raise ValueError(f"Cannot cast {self.dtype} to {dtype}")

        return LocalOperator(
            hilbert=self.hilbert,
            operators=[np.copy(op) for op in self._operators],
            acting_on=self._acting_on_list(),
            constant=self._constant,
            dtype=dtype,
        )

    def transpose(self, *, concrete=False):
        r"""LocalOperator: Returns the tranpose of this operator."""
        if concrete:

            new_ops = [np.copy(ops.transpose()) for ops in self._operators]

            return LocalOperator(
                hilbert=self.hilbert,
                operators=new_ops,
                acting_on=self._acting_on_list(),
                constant=self._constant,
            )
        else:
            return Transpose(self)

    def conjugate(self, *, concrete=False):
        r"""LocalOperator: Returns the complex conjugate of this operator."""
        new_ops = [np.copy(ops).conjugate() for ops in self._operators]

        return LocalOperator(
            hilbert=self.hilbert,
            operators=new_ops,
            acting_on=self._acting_on_list(),
            constant=np.conjugate(self._constant),
        )

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        max_size = self.n_operators if self._nonzero_diagonal else 0
        for op in self._operators:
            nnz_mat = np.abs(op) > self.mel_cutoff
            nnz_mat[np.diag_indices(nnz_mat.shape[0])] = False
            nnz_rows = np.sum(nnz_mat, axis=1)
            max_size += np.max(nnz_rows)

        return max_size

    def get_conn_flattened(self, x, sections, pad=False):
        r"""Finds the connected elements of the Operator. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.
            pad (bool): Whether to use zero-valued matrix elements in order to return all equal sections.

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """

        return self._get_conn_flattened_kernel(
            np.asarray(x),
            sections,
            self._local_states,
            self._basis,
            self._constant,
            self._diag_mels,
            self._n_conns,
            self._mels,
            self._x_prime,
            self._acting_on,
            self._acting_size,
            self._nonzero_diagonal,
            pad,
        )

    def _get_conn_flattened_closure(self):
        _local_states = self._local_states
        _basis = self._basis
        _constant = self._constant
        _diag_mels = self._diag_mels
        _n_conns = self._n_conns
        _mels = self._mels
        _x_prime = self._x_prime
        _acting_on = self._acting_on
        _acting_size = self._acting_size
        # workaround my painfully discovered Numba#6979 (cannot use numpy bools in closures)
        _nonzero_diagonal = bool(self._nonzero_diagonal)

        fun = self._get_conn_flattened_kernel

        def gccf_fun(x, sections):
            return fun(
                x,
                sections,
                _local_states,
                _basis,
                _constant,
                _diag_mels,
                _n_conns,
                _mels,
                _x_prime,
                _acting_on,
                _acting_size,
                _nonzero_diagonal,
            )

        return jit(nopython=True)(gccf_fun)

    @staticmethod
    @jit(nopython=True)
    def _get_conn_flattened_kernel(
        x,
        sections,
        local_states,
        basis,
        constant,
        diag_mels,
        n_conns,
        all_mels,
        all_x_prime,
        acting_on,
        acting_size,
        nonzero_diagonal,
        pad=False,
    ):
        batch_size = x.shape[0]
        n_sites = x.shape[1]
        dtype = all_mels.dtype

        assert sections.shape[0] == batch_size

        n_operators = n_conns.shape[0]
        xs_n = np.empty((batch_size, n_operators), dtype=np.intp)

        tot_conn = 0
        max_conn = 0

        for b in range(batch_size):
            # diagonal element
            conn_b = 1 if nonzero_diagonal else 0

            # counting the off-diagonal elements
            for i in range(n_operators):
                acting_size_i = acting_size[i]

                xs_n[b, i] = 0
                x_b = x[b]
                x_i = x_b[acting_on[i, :acting_size_i]]
                for k in range(acting_size_i):
                    xs_n[b, i] += (
                        np.searchsorted(
                            local_states[i, acting_size_i - k - 1],
                            x_i[acting_size_i - k - 1],
                        )
                        * basis[i, k]
                    )

                conn_b += n_conns[i, xs_n[b, i]]

            tot_conn += conn_b
            sections[b] = tot_conn

            if pad:
                max_conn = max(conn_b, max_conn)

        if pad:
            tot_conn = batch_size * max_conn

        x_prime = np.empty((tot_conn, n_sites), dtype=x.dtype)
        mels = np.empty(tot_conn, dtype=dtype)

        c = 0
        for b in range(batch_size):
            c_diag = c
            x_batch = x[b]

            if nonzero_diagonal:
                mels[c_diag] = constant
                x_prime[c_diag] = np.copy(x_batch)
                c += 1

            for i in range(n_operators):

                # Diagonal part
                #  If nonzero_diagonal, this goes to c_diag = 0 ....
                # if zero_diagonal this just sets the last element to 0
                # so it's not worth it skipping it
                mels[c_diag] += diag_mels[i, xs_n[b, i]]
                n_conn_i = n_conns[i, xs_n[b, i]]

                if n_conn_i > 0:
                    sites = acting_on[i]
                    acting_size_i = acting_size[i]

                    for cc in range(n_conn_i):
                        mels[c + cc] = all_mels[i, xs_n[b, i], cc]
                        x_prime[c + cc] = np.copy(x_batch)

                        for k in range(acting_size_i):
                            x_prime[c + cc, sites[k]] = all_x_prime[
                                i, xs_n[b, i], cc, k
                            ]
                    c += n_conn_i

            if pad:
                delta_conn = max_conn - (c - c_diag)
                mels[c : c + delta_conn].fill(0)
                x_prime[c : c + delta_conn, :] = np.copy(x_batch)
                c += delta_conn
                sections[b] = c

        return x_prime, mels

    def get_conn_filtered(self, x, sections, filters):
        r"""Finds the connected elements of the Operator using only a subset of operators. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.
            filters (array): Only operators op(filters[i]) are used to find the connected elements of
                        x[i].

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """

        return self._get_conn_filtered_kernel(
            x,
            sections,
            self._local_states,
            self._basis,
            self._constant,
            self._diag_mels,
            self._n_conns,
            self._mels,
            self._x_prime,
            self._acting_on,
            self._acting_size,
            filters,
        )

    @staticmethod
    @jit(nopython=True)
    def _get_conn_filtered_kernel(
        x,
        sections,
        local_states,
        basis,
        constant,
        diag_mels,
        n_conns,
        all_mels,
        all_x_prime,
        acting_on,
        acting_size,
        filters,
    ):

        batch_size = x.shape[0]
        n_sites = x.shape[1]
        dtype = all_mels.dtype

        assert filters.shape[0] == batch_size and sections.shape[0] == batch_size

        n_operators = n_conns.shape[0]
        xs_n = np.empty((batch_size, n_operators), dtype=np.intp)

        tot_conn = 0

        for b in range(batch_size):
            # diagonal element
            tot_conn += 1

            # counting the off-diagonal elements
            i = filters[b]

            assert i < n_operators and i >= 0
            acting_size_i = acting_size[i]

            xs_n[b, i] = 0
            x_b = x[b]
            x_i = x_b[acting_on[i, :acting_size_i]]
            for k in range(acting_size_i):
                xs_n[b, i] += (
                    np.searchsorted(
                        local_states[i, acting_size_i - k - 1],
                        x_i[acting_size_i - k - 1],
                    )
                    * basis[i, k]
                )

            tot_conn += n_conns[i, xs_n[b, i]]
            sections[b] = tot_conn

        x_prime = np.empty((tot_conn, n_sites))
        mels = np.empty(tot_conn, dtype=dtype)

        c = 0
        for b in range(batch_size):
            c_diag = c
            mels[c_diag] = constant
            x_batch = x[b]
            x_prime[c_diag] = np.copy(x_batch)
            c += 1

            i = filters[b]
            # Diagonal part
            mels[c_diag] += diag_mels[i, xs_n[b, i]]
            n_conn_i = n_conns[i, xs_n[b, i]]

            if n_conn_i > 0:
                sites = acting_on[i]
                acting_size_i = acting_size[i]

                for cc in range(n_conn_i):
                    mels[c + cc] = all_mels[i, xs_n[b, i], cc]
                    x_prime[c + cc] = np.copy(x_batch)

                    for k in range(acting_size_i):
                        x_prime[c + cc, sites[k]] = all_x_prime[i, xs_n[b, i], cc, k]
                c += n_conn_i

        return x_prime, mels

    def __repr__(self):
        ao = self.acting_on

        acting_str = f"acting_on={ao}"
        if len(acting_str) > 55:
            acting_str = f"#acting_on={len(ao)} locations"
        return f"{type(self).__name__}(dim={self.hilbert.size}, {acting_str}, constant={self.constant}, dtype={self.dtype})"



# Copyright 2020 The Netket Authors. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import networkx as nx
from netket import legacy as nk

# Sigma^z*Sigma^z interactions
sigmaz = [[1, 0], [0, -1]]
mszsz = np.kron(sigmaz, sigmaz)

# Exchange interactions
exchange = np.asarray([[0, 0, 0, 0], [0, 0, 2, 0], [0, 2, 0, 0], [0, 0, 0, 0]])

# Couplings J1 and J2
J = [1, 0.4]
L = 20

pars = {}

# Define bond operators, labels, and couplings
bond_operator = [
    (J[0] * mszsz).tolist(),
    (J[1] * mszsz).tolist(),
    (-J[0] * exchange).tolist(),
    (J[1] * exchange).tolist(),
]

bond_color = [1, 2, 1, 2]

# Define custom graph
G = nx.Graph()
for i in range(L):
    G.add_edge(i, (i + 1) % L, color=1)
    G.add_edge(i, (i + 2) % L, color=2)

edge_colors = [[u, v, G[u][v]["color"]] for u, v in G.edges]

# Custom Graph
g = nk.graph.CustomGraph(edge_colors)

# Spin based Hilbert Space
hi = nk.hilbert.Spin(s=1 / 2, total_sz=0.0, N=g.n_nodes)

# Custom Hamiltonian operator
op = nk.operator.GraphOperator(hi, bondops=bond_operator, bondops_colors=bond_color)

# Restricted Boltzmann Machine
ma = nk.machine.RbmSpin(hi, alpha=1)
ma.init_random_parameters(seed=1234, sigma=0.01)

# Sampler
sa = nk.sampler.MetropolisHamiltonianPt(machine=ma, hamiltonian=op, n_replicas=16)

# Optimizer
opt = nk.optimizer.Sgd(ma, learning_rate=0.01)

# Variational Monte Carlo
gs = nk.variational.Vmc(
    hamiltonian=op,
    sampler=sa,
    optimizer=opt,
    n_samples=1000,
    use_iterative=True,
    method="Sr",
)

gs.run(out="test", n_iter=10000)



from netket import legacy as nk
import torch
import numpy as np
import cProfile


class LogCoshSum(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self._ls = torch.nn.LogSigmoid()
        self._log2 = np.log(2.0)

    def forward(self, x):
        batch_size = x.shape[0]
        y = torch.zeros((batch_size, 2), dtype=torch.float64)
        x = torch.abs(x)
        y[:, 0] = (x - self._ls.forward(2 * x) - self._log2).sum(dim=1)
        return y


# 1D Lattice
g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)

# Ising spin hamiltonian
ha = nk.operator.Ising(hilbert=hi, graph=g, h=1.0)

input_size = hi.size
alpha = 1

model = torch.nn.Sequential(
    torch.nn.Linear(input_size, alpha * input_size),
    LogCoshSum(),
)

ma = nk.machine.Torch(model, hilbert=hi)

ma.parameters = 0.1 * (np.random.randn(ma.n_par))

# Metropolis Local Sampling
sa = nk.sampler.MetropolisLocal(machine=ma, n_chains=64)

# Optimizer
op = nk.optimizer.Sgd(ma, 0.1)

gs = nk.Vmc(
    hamiltonian=ha,
    sampler=sa,
    optimizer=op,
    n_samples=1000,
    n_discard=2,
    sr=nk.optimizer.SR(ma, diag_shift=0.1),
)

gs.run(out="test", n_iter=300)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Callable

import jax
from jax import numpy as jnp
from jax.experimental import loops

from netket.hilbert import AbstractParticle

from netket.utils import mpi
from netket.utils.types import PyTree, PRNGKeyT

from netket.utils.deprecation import deprecated, warn_deprecation
from netket.utils import struct

from .base import Sampler, SamplerState


@struct.dataclass
class MetropolisRule:
    """
    Base class for Transition rules of Metropolis, such as Local, Exchange, Hamiltonian
    and several others.
    """

    def init_state(
        rule,
        sampler: Sampler,
        machine: Callable,
        params: PyTree,
        key: PRNGKeyT,
    ) -> Optional[Any]:
        """
        Initialises the optional internal state of the Metropolis Sampler Transition
        Rule.

        The provided key is unique and does not need to be splitted.
        It should return an immutable datastructure.

        Arguments:
            sampler: The Metropolis sampler
            machine: The forward evaluation function of the model, accepting PyTrees of parameters and inputs.
            params: The dict of variables needed to evaluate the model.
            key: A Jax PRNGKey rng state.

        Returns:
            An Optional State.
        """
        return None

    def reset(
        rule,
        sampler: Sampler,
        machine: Callable,
        params: PyTree,
        sampler_state: SamplerState,
    ) -> Optional[Any]:
        """
        Resets the internal state of the Metropolis Sampler Transition Rule.

        Arguments:
            sampler: The Metropolis sampler
            machine: The forward evaluation function of the model, accepting PyTrees of parameters and inputs.
            params: The dict of variables needed to evaluate the model.
            sampler_state: The current state of the sampler. Should not modify it.

        Returns:
           A new, resetted, state of the rule. This returns the same type of :py:meth:`sampler_state.rule_state` and might be None.

        """
        return sampler_state.rule_state

    def transition(
        rule,
        sampler: Sampler,
        machine: Callable,
        parameters: PyTree,
        state: SamplerState,
        key: PRNGKeyT,
        σ: jnp.ndarray,
    ) -> jnp.ndarray:

        raise NotImplementedError("error")

    def random_state(
        rule,
        sampler: Sampler,
        machine: Callable,
        parameters: PyTree,
        state: SamplerState,
        key: PRNGKeyT,
    ):
        """
        Generates a random state compatible with this rule.

        By default this calls :func:`netket.hilbert.random.random_state`.

        Arguments:
            sampler: the sampler
            machine: the function to evaluate the model
            parameters: the parameters of the model
            state: the current sampler state
            key: the PRNGKey to use to generate the random state
        """
        return sampler.hilbert.random_state(
            key, size=sampler.n_batches, dtype=sampler.dtype
        )


@struct.dataclass
class MetropolisSamplerState(SamplerState):
    """
    State for a metropolis sampler.
    Contains the current configuration, the rng state and the (optional)
    state of the transition rule.
    """

    σ: jnp.ndarray
    """Current batch of configurations in the markov chain."""
    rng: jnp.ndarray
    """State of the random number generator (key, in jax terms)."""
    rule_state: Optional[Any]
    """Optional state of a transition rule."""
    n_steps_proc: int = 0
    """Number of moves performed along the chains in this process since the last reset."""
    n_accepted_proc: int = 0
    """Number of accepted transitions among the chains in this process since the last reset."""

    @property
    def acceptance(self) -> float:
        """The fraction of accepted moves across all chains and MPI processes.

        The rate is computed since the last reset of the sampler.
        Will return None if no sampling has been performed since then.
        """
        if self.n_steps == 0:
            return None

        return self.n_accepted / self.n_steps

    @property
    @deprecated(
        """Please use the attribute `.acceptance` instead of
        `.acceptance_ratio`. The new attribute `.acceptance` returns the
        acceptance ratio ∈ [0,1], instead of the current `acceptance_ratio`
        returning a percentage, which is a bug."""
    )
    def acceptance_ratio(self):
        """DEPRECATED: Please use the attribute `.acceptance` instead of
        `.acceptance_ratio`. The new attribute `.acceptance` returns the
        acceptance ratio ∈ [0,1], instead of the current `acceptance_ratio`
        returning a percentage, which is a bug.

        The percentage of accepted moves across all chains and MPI processes.

        The rate is computed since the last reset of the sampler.
        Will return None if no sampling has been performed since then.
        """
        return self.acceptance * 100

    @property
    def n_steps(self) -> int:
        """Total number of moves performed across all processes since the last reset."""
        return self.n_steps_proc * mpi.n_nodes

    @property
    def n_accepted(self) -> int:
        """Total number of moves accepted across all processes since the last reset."""
        res, _ = mpi.mpi_sum_jax(self.n_accepted_proc)
        return res

    def __repr__(self):
        if self.n_steps > 0:
            acc_string = "# accepted = {}/{} ({}%), ".format(
                self.n_accepted, self.n_steps, self.acceptance * 100
            )
        else:
            acc_string = ""

        return f"{type(self).__name__}({acc_string}rng state={self.rng})"


@struct.dataclass
class MetropolisSampler(Sampler):
    r"""
    Metropolis-Hastings sampler for an Hilbert space according to a specific transition rule.

    The transition rule is used to generate a proposed state :math:`s^\prime`, starting from the
    current state :math:`s`. The move is accepted with probability

    .. math::

        A(s \rightarrow s^\prime) = \mathrm{min} \left( 1,\frac{P(s^\prime)}{P(s)} F(e^{L(s,s^\prime)}) \right) ,

    where the probability being sampled from is :math:`P(s)=|M(s)|^p. Here ::math::`M(s)` is a
    user-provided function (the machine), :math:`p` is also user-provided with default value :math:`p=2`,
    and :math:`L(s,s^\prime)` is a suitable correcting factor computed by the transition kernel.

    The dtype of the sampled states can be chosen.
    """

    rule: MetropolisRule = None
    """The metropolis transition rule."""
    n_sweeps: int = struct.field(pytree_node=False, default=None)
    """Number of sweeps for each step along the chain. Defaults to number of sites in hilbert space."""
    reset_chains: bool = struct.field(pytree_node=False, default=False)
    """If True resets the chain state when reset is called (every new sampling)."""

    def __pre_init__(self, hilbert, rule, **kwargs):
        r"""
        Constructs a Metropolis Sampler.

        Args:
            hilbert: The hilbert space to sample
            rule: A `MetropolisRule` to generate random transitions from a given state as
                    well as uniform random states.
            n_sweeps: The number of exchanges that compose a single sweep.
                    If None, sweep_size is equal to the number of degrees of freedom being sampled
                    (the size of the input vector s to the machine).
            reset_chains: If False the state configuration is not resetted when reset() is called.
            n_chains: The number of Markov Chain to be run in parallel on a single process.
            machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
            dtype: The dtype of the statees sampled (default = np.float32).
        """
        # process arguments in the base
        args, kwargs = super().__pre_init__(hilbert=hilbert, **kwargs)

        kwargs["rule"] = rule

        # deprecation warnings
        if "reset_chain" in kwargs:
            warn_deprecation(
                "The keyword argument `reset_chain` is deprecated in favour of `reset_chains`"
            )
            kwargs["reset_chains"] = kwargs.pop("reset_chain")

        return args, kwargs

    def __post_init__(self):
        super().__post_init__()
        # Validate the inputs
        if not isinstance(self.rule, MetropolisRule):
            raise TypeError("rule must be a MetropolisRule.")

        if not isinstance(self.reset_chains, bool):
            raise TypeError("reset_chains must be a boolean.")

        #  Default value of n_sweeps
        if self.n_sweeps is None:
            object.__setattr__(self, "n_sweeps", self.hilbert.size)

    def _init_state(sampler, machine, params, key):
        key_state, key_rule = jax.random.split(key, 2)
        rule_state = sampler.rule.init_state(sampler, machine, params, key_rule)
        σ = jnp.zeros(
            (sampler.n_chains_per_rank, sampler.hilbert.size), dtype=sampler.dtype
        )

        state = MetropolisSamplerState(σ=σ, rng=key_state, rule_state=rule_state)

        # If we don't reset the chain at every sampling iteration, then reset it
        # now.
        if not sampler.reset_chains:
            key_state, rng = jax.random.split(key_state)
            σ = sampler.rule.random_state(sampler, machine, params, state, rng)
            state = state.replace(σ=σ, rng=key_state)

        return state

    def _reset(sampler, machine, parameters, state):
        new_rng, rng = jax.random.split(state.rng)

        if sampler.reset_chains:
            σ = sampler.rule.random_state(sampler, machine, parameters, state, rng)
        else:
            σ = state.σ

        rule_state = sampler.rule.reset(sampler, machine, parameters, state)

        return state.replace(
            σ=σ, rng=new_rng, rule_state=rule_state, n_steps_proc=0, n_accepted_proc=0
        )

    def _sample_next(sampler, machine, parameters, state):
        new_rng, rng = jax.random.split(state.rng)

        with loops.Scope() as s:
            s.key = rng
            s.σ = state.σ
            s.log_prob = sampler.machine_pow * machine.apply(parameters, state.σ).real

            # for logging
            s.accepted = state.n_accepted_proc

            for i in s.range(sampler.n_sweeps):
                # 1 to propagate for next iteration, 1 for uniform rng and n_chains for transition kernel
                s.key, key1, key2 = jax.random.split(s.key, 3)

                σp, log_prob_correction = sampler.rule.transition(
                    sampler, machine, parameters, state, key1, s.σ
                )
                proposal_log_prob = (
                    sampler.machine_pow * machine.apply(parameters, σp).real
                )

                uniform = jax.random.uniform(key2, shape=(sampler.n_chains_per_rank,))
                if log_prob_correction is not None:
                    do_accept = uniform < jnp.exp(
                        proposal_log_prob - s.log_prob + log_prob_correction
                    )
                else:
                    do_accept = uniform < jnp.exp(proposal_log_prob - s.log_prob)

                # do_accept must match ndim of proposal and state (which is 2)
                s.σ = jnp.where(do_accept.reshape(-1, 1), σp, s.σ)
                s.accepted += do_accept.sum()

                s.log_prob = jax.numpy.where(
                    do_accept.reshape(-1), proposal_log_prob, s.log_prob
                )

            new_state = state.replace(
                rng=new_rng,
                σ=s.σ,
                n_accepted_proc=s.accepted,
                n_steps_proc=state.n_steps_proc
                + sampler.n_sweeps * sampler.n_chains_per_rank,
            )

        return new_state, new_state.σ

    def __repr__(sampler):
        return (
            f"{type(sampler).__name__}("
            + "\n  hilbert = {},".format(sampler.hilbert)
            + "\n  rule = {},".format(sampler.rule)
            + "\n  n_chains = {},".format(sampler.n_chains)
            + "\n  machine_power = {},".format(sampler.machine_pow)
            + "\n  reset_chains = {},".format(sampler.reset_chains)
            + "\n  n_sweeps = {},".format(sampler.n_sweeps)
            + "\n  dtype = {}".format(sampler.dtype)
            + ")"
        )

    def __str__(sampler):
        return (
            f"{type(sampler).__name__}("
            + "rule = {}, ".format(sampler.rule)
            + "n_chains = {}, ".format(sampler.n_chains)
            + "machine_power = {}, ".format(sampler.machine_pow)
            + "n_sweeps = {}, ".format(sampler.n_sweeps)
            + "dtype = {})".format(sampler.dtype)
        )


def MetropolisLocal(hilbert, *args, **kwargs) -> MetropolisSampler:
    r"""
    Sampler acting on one local degree of freedom.

    This sampler acts locally only on one local degree of freedom :math:`s_i`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s_N`,
    where :math:`s^\prime_i \neq s_i`.

    The transition probability associated to this
    sampler can be decomposed into two steps:

    1. One of the site indices :math:`i = 1\dots N` is chosen with uniform probability.

    2. Among all the possible (:math:`m`) values that :math:`s_i` can take,
    one of them is chosen with uniform probability.

    For example, in the case of spin :math:`1/2` particles, :math:`m=2`
    and the possible local values are :math:`s_i = -1,+1`.
    In this case then :class:`MetropolisLocal` is equivalent to flipping a random spin.

    In the case of bosons, with occupation numbers
    :math:`s_i = 0, 1, \dots n_{\mathrm{max}}`, :class:`MetropolisLocal`
    would pick a random local occupation number uniformly between :math:`0`
    and :math:`n_{\mathrm{max}}`.

    Args:
        hilbert: The hilbert space to sample
        n_chains: The number of Markov Chain to be run in parallel on a single process.
        n_sweeps: The number of exchanges that compose a single sweep.
                If None, sweep_size is equal to the number of degrees of freedom being sampled
                (the size of the input vector s to the machine).
        n_chains: The number of batches of the states to sample (default = 8)
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the statees sampled (default = np.float32).
    """
    from .rules import LocalRule

    return MetropolisSampler(hilbert, LocalRule(), *args, **kwargs)


def MetropolisExchange(
    hilbert, *args, clusters=None, graph=None, d_max=1, **kwargs
) -> MetropolisSampler:
    r"""
    This sampler acts locally only on two local degree of freedom :math:`s_i` and :math:`s_j`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s^\prime_j \dots s_N`,
    where in general :math:`s^\prime_i \neq s_i` and :math:`s^\prime_j \neq s_j`.
    The sites :math:`i` and :math:`j` are also chosen to be within a maximum graph
    distance of :math:`d_{\mathrm{max}}`.

    The transition probability associated to this sampler can
    be decomposed into two steps:

    1. A pair of indices :math:`i,j = 1\dots N`, and such that
       :math:`\mathrm{dist}(i,j) \leq d_{\mathrm{max}}`,
       is chosen with uniform probability.

    2. The sites are exchanged, i.e. :math:`s^\prime_i = s_j` and :math:`s^\prime_j = s_i`.


    Notice that this sampling method generates random permutations of the quantum
    numbers, thus global quantities such as the sum of the local quantum numbers
    are conserved during the sampling.
    This scheme should be used then only when sampling in a
    region where :math:`\sum_i s_i = \mathrm{constant} ` is needed,
    otherwise the sampling would be strongly not ergodic.

    Args:
        hilbert: The hilbert space to sample
        d_max: The maximum graph distance allowed for exchanges.
        n_chains: The number of Markov Chain to be run in parallel on a single process.
        n_sweeps: The number of exchanges that compose a single sweep.
                If None, sweep_size is equal to the number of degrees of freedom being sampled
                (the size of the input vector s to the machine).
        n_batches: The number of batches of the states to sample (default = 8)
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the statees sampled (default = np.float32).


    Examples:
          Sampling from a RBM machine in a 1D lattice of spin 1/2, using
          nearest-neighbor exchanges.

          >>> import netket as nk
          >>>
          >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
          >>> hi=nk.hilbert.Spin(s=0.5, N=g.n_nodes)
          >>>
          >>> # Construct a MetropolisExchange Sampler
          >>> sa = nk.sampler.MetropolisExchange(hi, graph=g)
          >>> print(sa)
          MetropolisSampler(rule = ExchangeRule(# of clusters: 200), n_chains = 16, machine_power = 2, n_sweeps = 100, dtype = <class 'numpy.float64'>)
    """
    from .rules import ExchangeRule

    rule = ExchangeRule(clusters=clusters, graph=graph, d_max=d_max)
    return MetropolisSampler(hilbert, rule, *args, **kwargs)


def MetropolisHamiltonian(hilbert, hamiltonian, *args, **kwargs) -> MetropolisSampler:
    r"""
    Sampling based on the off-diagonal elements of a Hamiltonian (or a generic Operator).
    In this case, the transition matrix is taken to be:

    .. math::
       T( \mathbf{s} \rightarrow \mathbf{s}^\prime) = \frac{1}{\mathcal{N}(\mathbf{s})}\theta(|H_{\mathbf{s},\mathbf{s}^\prime}|),

    where :math:`\theta(x)` is the Heaviside step function, and :math:`\mathcal{N}(\mathbf{s})`
    is a state-dependent normalization.
    The effect of this transition probability is then to connect (with uniform probability)
    a given state :math:`\mathbf{s}` to all those states :math:`\mathbf{s}^\prime` for which the Hamiltonian has
    finite matrix elements.
    Notice that this sampler preserves by construction all the symmetries
    of the Hamiltonian. This is in generally not true for the local samplers instead.

    This sampler only works on the CPU. To use the Hamiltonian smapler with GPUs,
    you should use :class:`netket.sampler.MetropolisHamiltonianNumpy`

    Args:
       machine: A machine :math:`\Psi(s)` used for the sampling.
                The probability distribution being sampled
                from is :math:`F(\Psi(s))`, where the function
                :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
       hamiltonian: The operator used to perform off-diagonal transition.
       n_chains: The number of Markov Chain to be run in parallel on a single process.
       sweep_size: The number of exchanges that compose a single sweep.
                   If None, sweep_size is equal to the number of degrees of freedom (n_visible).


    Examples:
       Sampling from a RBM machine in a 1D lattice of spin 1/2

       >>> import netket as nk
       >>>
       >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
       >>> hi=nk.hilbert.Spin(s=0.5, N=g.n_nodes)
       >>>
       >>> # Transverse-field Ising Hamiltonian
       >>> ha = nk.operator.Ising(hilbert=hi, h=1.0, graph=g)
       >>>
       >>> # Construct a MetropolisExchange Sampler
       >>> sa = nk.sampler.MetropolisHamiltonian(hi, hamiltonian=ha)
       >>> print(sa)
       MetropolisSampler(rule = HamiltonianRule(Ising(J=1.0, h=1.0; dim=100)), n_chains = 16, machine_power = 2, n_sweeps = 100, dtype = <class 'numpy.float64'>)
    """
    from .rules import HamiltonianRule

    rule = HamiltonianRule(hamiltonian)
    return MetropolisSampler(hilbert, rule, *args, **kwargs)


def MetropolisGaussian(hilbert, sigma=1.0, *args, **kwargs) -> MetropolisSampler:
    r"""This sampler acts on all particle positions simultaneously
    and proposes a new state according to a Gaussian distribution
    with width sigma.
    Args:
       hilber: The continuous Hilbert space
       sigma: The width of the Gaussian proposal distribution (default: 1.0)
       n_chains: The number of Markov Chain to be run in parallel on a single process.
       sweep_size: The number of exchanges that compose a single sweep.
                   If None, sweep_size is equal to the number of degrees of freedom (n_visible).
    """
    if not isinstance(hilbert, AbstractParticle):
        raise ValueError("This sampler only works for Continuous Hilbert spaces.")

    from .rules import GaussianRule

    rule = GaussianRule(sigma)
    return MetropolisSampler(hilbert, rule, *args, **kwargs)



import netket as nk
import time

SEED = 3141592
L = 8


def _vmc(n_iter=20):
    hi = nk.hilbert.Spin(s=0.5) ** L

    ma = nk.models.RBM(alpha=1)

    ha = nk.operator.Ising(hi, nk.graph.Hypercube(length=L, n_dim=1), h=1.0)
    sa = nk.sampler.MetropolisLocal(hi)
    vs = nk.vqs.MCState(sa, ma, n_samples=500, seed=SEED)

    op = nk.optimizer.Sgd(learning_rate=0.1)

    return nk.Vmc(hamiltonian=ha, variational_state=vs, optimizer=op)


def test_earlystopping_with_patience():
    patience = 10
    es = nk.callbacks.EarlyStopping(patience=patience)
    es._best_val = -1e6
    vmc = _vmc()

    vmc.run(20, callback=es)

    assert vmc.step_count == patience


def test_earlystopping_with_baseline():
    baseline = -10
    es = nk.callbacks.EarlyStopping(baseline=baseline)
    vmc = _vmc()

    vmc.run(20, callback=es)
    # We should actually assert something..


def test_timeout():
    timeout = 5
    tout = nk.callbacks.Timeout(timeout=timeout)
    vmc = _vmc()

    # warmup the jit
    vmc.run(1)

    st = time.time()
    vmc.run(20000, callback=tout)
    runtime = time.time() - st

    # There is a lag in the first iteration of about 3 seconds
    # But the timeout works!
    assert abs(timeout - runtime) < 3
